import __builtin__, imp, os, sqlite3, sys, types

from lib2to3.pgen2 import tokenize
from lib2to3.refactor import get_fixers_from_package, RefactoringTool

database = "traces.db"
conn = sqlite3.connect(database)


tracing_functions = []
def register_tracing_function(f):
  tracing_functions.append(f)
  return f


def create_table(table, *columns):
  with conn:
    v = ', '.join(columns)
    conn.execute(
      "CREATE TABLE IF NOT EXISTS %s (%s, UNIQUE (%s))" % (table, v, v)
    )

  def insert_unique(*values):
    with conn:
      try:
        conn.execute(
          'INSERT INTO %s VALUES (%s)' % (table, ', '.join('?' * len(values))),
          values
        )
      except sqlite3.IntegrityError as e:
        if not str(e).startswith('UNIQUE constraint failed:'):
          raise
  return insert_unique


def get_fixers():
  return [
    f for f in get_fixers_from_package("my2to3.fixes") if f.endswith("_trace")
  ]


def get_data(table, columns_to_select='*', conditions={}):
  query = "SELECT %s FROM %s" % (
    ', '.join(columns_to_select),
    table,
  )
  if conditions:
    query += " WHERE " + ' AND '.join(k + " = :" + k for k in conditions.keys())
  with conn:
    return conn.execute(query, conditions).fetchall()


def apply_fixers(string, name):
  # This function is inspired by refactoring_tool.refactor_file

  refactoring_tool = RefactoringTool(fixer_names=get_fixers())

  s = string + "\n" # Silence certain parse errors
  encoding = tokenize.detect_encoding(lambda: s)[0]
  s = s.decode(encoding)
  s = unicode(refactoring_tool.refactor_string(s, name))
  # The [:-1] is to take off the \n we added earlier
  return s[:-1].encode(encoding)


init_py = '__init__' + os.extsep + 'py'
class ModuleImporter:
  def __init__(self, is_whitelisted):
    self.paths = {}
    self.is_whitelisted = is_whitelisted

  def find_module(self, fullname, path=None):
    if not path or not self.is_whitelisted(fullname, path):
      return

    try:
      file, pathname, description = imp.find_module(fullname.rsplit('.', 1)[-1], path)
    except ImportError:
      return
    if file is not None:
      file.close()

    if description[-1] is imp.PKG_DIRECTORY:
      # It's a package
      assert file is None, file
      pathname = os.path.join(pathname, init_py)

    self.paths[fullname] = pathname
    return self

  def load_module(self, fullname):
    imp.acquire_lock()
    try:
      if fullname in sys.modules:
        return sys.modules[fullname]

      path = self.paths[fullname]
      with open(path) as f:
        new_code = compile(apply_fixers(f.read(), path), path, 'exec')

      sys.modules[fullname] = module = types.ModuleType(fullname)
      module.__file__ = p = new_code.co_filename
      if p.endswith(init_py):
        module.__path__ = [path.rstrip(init_py)]
      try:
        exec new_code in module.__dict__
      except Exception:
        del sys.modules[fullname]
        raise
    finally:
      imp.release_lock()
    return module


def patch_imports(whitelist):
  sys.meta_path.append(ModuleImporter(whitelist))

  for f in tracing_functions:
    setattr(__builtin__, f.__name__, f)


for fixer in get_fixers():
  # Import each fixer, to make sure that "my2to3.trace.tracing_functions" is
  # correctly populated.
  __import__(fixer)
